@@ -135,54 +135,71 @@ class SqliteBackedCorpus(BaseCorpus):
 
 
 class PlainTextSqliteCorpus(SqliteBackedCorpus):
-
     CORPUS_TYPE = "PlainTextSqliteCorpus"
 
+    def __init__(self, db_path, tokeniser=utilities.tokens):
+        super().__init__(db_path)
+
+        self.tokeniser = tokeniser
+
+    def __getstate__(self):
+        return self.db_path, self.tokeniser
+
+    def __setstate__(self, state):
+        self.__init__(state[0], tokeniser=state[1])
+
     def replace_docs(self, texts):
         """Replace the existing documents with texts."""
         self.db.execute("pragma journal_mode=WAL")
         self.db.execute("savepoint add_texts")
 
-        self.db.execute("drop table if exists doc")
-        self.db.execute(
-            """
-            create table doc(
-                doc_id integer primary key,
-                text not null
+        try:
+            self.db.execute("drop table if exists doc")
+            self.db.execute(
+                """
+                create table doc(
+                    doc_id integer primary key,
+                    text not null
+                )
+                """
             )
-            """
-        )
 
-        self.db.execute("delete from doc")
-        self.db.executemany(
-            "insert or ignore into doc(text) values(?)", ([t] for t in texts)
-        )
+            self.db.execute("delete from doc")
+            self.db.executemany(
+                "insert or ignore into doc(text) values(?)", ([t] for t in texts)
+            )
 
-        self.db.execute("release add_texts")
+        except Exception:
+            self.db.execute("rollback to add_texts")
+
+        finally:
+            self.db.execute("release add_texts")
 
     def docs(self, doc_keys=None):
-
         self.db.execute("savepoint docs")
+        try:
+            # Note that it's valid to pass an empty sequence of doc_keys,
+            # so we need to check sentinel explicitly.
+            if doc_keys is None:
+                doc_keys = self.keys()
 
-        # Note that it's valid to pass an empty sequence of doc_keys,
-        # so we need to check sentinel explicitly.
-        if doc_keys is None:
-            doc_keys = self.keys()
+            for key in doc_keys:
+                doc = list(
+                    self.db.execute(
+                        "select doc_id, text from doc where doc_id = ?", [key]
+                    )
+                )[0]
+                yield key, doc
 
-        for key in doc_keys:
-            doc = list(
-                self.db.execute("select doc_id, text from doc where doc_id = ?", [key])
-            )[0]
-            yield key, doc
-
-        self.db.execute("release docs")
+        finally:
+            self.db.execute("release docs")
 
     def keys(self):
         return (r["doc_id"] for r in self.db.execute("select doc_id from doc"))
 
     def index(self, doc):
         return {
-            "text": utilities.tokens(doc["text"]),
+            "text": self.tokeniser(doc["text"]),
         }
 
     def render_docs_html(self, doc_keys):
@@ -191,7 +208,6 @@ class PlainTextSqliteCorpus(SqliteBackedCorpus):
 
 
 class StackExchangeCorpus(SqliteBackedCorpus):
-
     CORPUS_TYPE = "StackExchangeCorpus"
 
     def add_site_data(self, posts_file, comments_file, users_file, site_url):
@@ -303,7 +319,6 @@ class StackExchangeCorpus(SqliteBackedCorpus):
             post_types = {"1": "Question", "2": "Answer"}
 
             for event, elem in tree:
-
                 # We only consider questions and answers - SX uses other post types
                 # to describe wiki's, tags, moderator nominations and more.
                 if elem.attrib.get("PostTypeId") not in ("1", "2"):
@@ -354,7 +369,6 @@ class StackExchangeCorpus(SqliteBackedCorpus):
             tree = ElementTree.iterparse(comments_file, events=("end",))
 
             for event, elem in tree:
-
                 doc = defaultdict(lambda: None)
                 doc.update(elem.attrib)
                 doc["site_id"] = site_id
@@ -379,7 +393,6 @@ class StackExchangeCorpus(SqliteBackedCorpus):
             tree = ElementTree.iterparse(users_file, events=("end",))
 
             for event, elem in tree:
-
                 doc = defaultdict(lambda: None)
                 doc.update(elem.attrib)
                 doc["site_id"] = site_id
@@ -423,7 +436,6 @@ class StackExchangeCorpus(SqliteBackedCorpus):
                 doc_keys = self.keys()
 
             for key in doc_keys:
-
                 doc = list(
                     self.db.execute(
                         """
@@ -546,7 +558,6 @@ class StackExchangeCorpus(SqliteBackedCorpus):
     )
 
     def _render_doc_key(self, key):
-
         base_fields = list(
             self.db.execute(
                 """
@@ -631,7 +642,6 @@ class StackExchangeCorpus(SqliteBackedCorpus):
         return (r["doc_id"] for r in self.db.execute("select doc_id from Post"))
 
     def index(self, doc):
-
         indexed = {
             "UserPosting": set([doc["DisplayName"]]),
             # Note tokenise different components separately, so there are
@@ -709,6 +719,27 @@ class TwittersphereCorpus(SqliteBackedCorpus):
 
     CORPUS_TYPE = "TwittersphereCorpus"
 
+    def __init__(self, db_path):
+        super().__init__(db_path)
+
+        version_rows = list(
+            self.db.execute(
+                "select value from metadata where key = 'twittersphere_schema_version'"
+            )
+        )
+
+        if not version_rows:
+            raise ValueError(
+                "This does not appear to be a Twittersphere generated database."
+            )
+
+        elif version_rows[0]["value"] < 15:
+            raise ValueError(
+                "The database needs to have a schema version of 15 or later. "
+                "You will need to regenerate the database with a newer "
+                "version of twittersphere to use this tool."
+            )
+
     def docs(self, doc_keys=None):
         self.db.execute("savepoint docs")
 
@@ -719,7 +750,6 @@ class TwittersphereCorpus(SqliteBackedCorpus):
                 doc_keys = self.keys()
 
             for tweet_id in doc_keys:
-
                 doc = list(
                     self.db.execute(
                         """
@@ -732,6 +762,23 @@ class TwittersphereCorpus(SqliteBackedCorpus):
                     )
                 )[0]
 
+                # Augment the doc with the specific collection context in which this tweet was
+                # directly collected. Since this might be multiple contexts this is a set.
+                doc["collection_labels"] = {
+                    r["label"]
+                    for r in self.db.execute(
+                        """
+                    select
+                        collection_context.label
+                    from tweet_at_time
+                    inner join collection_context using(context_id)
+                    where tweet_id = ?
+                        and directly_collected = 1
+                    """,
+                        [tweet_id],
+                    )
+                }
+
                 doc["hashtags"] = {
                     r["tag"]
                     for r in self.db.execute(
@@ -813,7 +860,6 @@ class TwittersphereCorpus(SqliteBackedCorpus):
             self.db.execute("release docs")
 
     def _retrieve_user(self, user_id):
-
         return list(
             self.db.execute("select * from user_latest where user_id = ?", [user_id])
         )[0]
@@ -867,7 +913,6 @@ class TwittersphereCorpus(SqliteBackedCorpus):
         )
 
     def index(self, doc):
-
         # Round down datetime to various degrees of granularity
         created_at = isoparse(doc["created_at"])
         rounded_times = utilities.round_datetime(created_at)
@@ -891,6 +936,7 @@ class TwittersphereCorpus(SqliteBackedCorpus):
                 "replied_to_user": [
                     doc.get("replied_to_user", {}).get("username", None)
                 ],
+                "collection_label": doc["collection_labels"],
             }
         else:
             indexed = {
